/*
	normalmap GIMP plugin

	Copyright (C) 2002 Shawn Kirst <skirst@fuse.net>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; see the file COPYING.  If not, write to
	the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA 02111-1307, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimp/gimpintl.h>

#include "preview3d.h"

enum FILTER_TYPE
{
   FILTER_NONE=0,FILTER_SOBEL,FILTER_PREWITT,FILTER_3x3,FILTER_5x5,
   FILTER_7x7,FILTER_9x9,
   MAX_FILTER_TYPE
};

enum CONVERSION_TYPE
{
   CONVERT_NONE=0,CONVERT_BIASED_RGB,CONVERT_RED,CONVERT_GREEN,CONVERT_BLUE,
   CONVERT_MAX_RGB,CONVERT_COLORSPACE,CONVERT_NORMALIZE_ONLY,
   MAX_CONVERSION_TYPE
};

enum DUDV_TYPE
{
   DUDV_NONE,DUDV_8BIT_SIGNED,DUDV_8BIT_UNSIGNED,DUDV_16BIT_SIGNED,
   DUDV_16BIT_UNSIGNED,
   MAX_DUDV_TYPE
};

typedef struct
{
	gint filter;
   gdouble scale;
   gint wrap;
   gint height_source;
   gint alpha;
   gint conversion;
   gint dudv;
} NormalmapVals;

static void query(void);
static void run(gchar *name, gint params, GimpParam *param,
					 gint *nreturn_vals, GimpParam **return_vals);

static gint32 normalmap(GimpDrawable *drawable, gboolean preview_mode);

static void ok_callback(GtkWidget *widget, gpointer data);
static gint normalmap_dialog(GimpDrawable *drawable);

GimpPlugInInfo PLUG_IN_INFO=
{
	NULL,NULL,query,run
};

NormalmapVals nmapvals=
{
	0,1,0,0,0,0,0
};

const float oneover255=1.0f/255.0f;

gint runme=0;

#define PREVIEW_SIZE  128
struct preview
{
   GtkWidget *widget;
   guchar *cache;
   gint width;
   gint height;
   gint bpp;
   gint rowstride;
};

static GtkWidget *dialog;
static struct preview preview;

MAIN()
	
static void query(void)
{
	static GimpParamDef args[]=
	{
		{GIMP_PDB_INT32,"run_mode","Interactive, non-interactive"},
		{GIMP_PDB_IMAGE,"image","Input image (unused)"},
		{GIMP_PDB_DRAWABLE,"drawable","Input drawable"},
		{GIMP_PDB_INT32,"filter","Filter type (0 = 4 sample, 1 = sobel, 2-6 = 3x3,5x5,7x7,9x9)"},
      {GIMP_PDB_FLOAT,"scale","Scale (>0)"},
      {GIMP_PDB_INT32,"wrap","Wrap (0 = no)"},
      {GIMP_PDB_INT32,"height_source","Height source (0 = average RGB, 1 = alpha channel)"},
      {GIMP_PDB_INT32,"alpha","Alpha (0 = unchanged, 1 = set to height, 2 = set to 0, 3 = set to 1)"},
      {GIMP_PDB_INT32,"conversion","Conversion (0 = normalize only, 1 = Biased RGB, 2 = Red, 3 = Green, 4 = Blue, 5 = Max RGB, 6 = Colorspace"},
      {GIMP_PDB_INT32,"dudv","DU/DV map (0 = none, 1 = 8-bit, 2 = 16-bit)"}      
	};
	static gint nargs=sizeof(args)/sizeof(args[0]);
	
	gimp_install_procedure("plug_in_normalmap",
								  "Converts image to an RGB normalmap",
								  "foo!",
								  "Shawn Kirst",
								  "Shawn Kirst",
								  "February 2002",
								  N_("<Image>/Filters/Map/Normalmap..."),
								  "RGB*",
								  GIMP_PLUGIN,
								  nargs,0,
								  args,NULL);
}

static void run(gchar *name, gint nparams, GimpParam *param,
					 gint *nreturn_vals, GimpParam **return_vals)
{
	static GimpParam values[1];
	GimpDrawable *drawable;
	GimpRunModeType run_mode;
	GimpPDBStatusType status=GIMP_PDB_SUCCESS;
	
	run_mode=param[0].data.d_int32;
	
	*nreturn_vals=1;
	*return_vals=values;
	
	values[0].type=GIMP_PDB_STATUS;
	values[0].data.d_status=status;
	
	drawable=gimp_drawable_get(param[2].data.d_drawable);
	
	switch(run_mode)
	{
		case GIMP_RUN_INTERACTIVE:
		   gimp_ui_init("normalmap",0);
		   gimp_get_data("plug_in_normalmap",&nmapvals);
		   if(!normalmap_dialog(drawable))
		   {
				gimp_drawable_detach(drawable);
				return;
			}
         g_free(preview.cache);
		   break;
		case GIMP_RUN_NONINTERACTIVE:
		   if(nparams!=5)
				status=GIMP_PDB_CALLING_ERROR;
		   else
		   {
			   nmapvals.filter=param[3].data.d_int32;
            nmapvals.scale=param[4].data.d_float;
            nmapvals.wrap=param[5].data.d_int32;
            nmapvals.height_source=param[6].data.d_int32;
            nmapvals.alpha=param[7].data.d_int32;
            nmapvals.conversion=param[8].data.d_int32;
            nmapvals.dudv=param[9].data.d_int32;
			}
		   break;
		case GIMP_RUN_WITH_LAST_VALS:
		   gimp_get_data("plug_in_normalmap",&nmapvals);
		   break;
		default:
		   break;
	}
	
	gimp_progress_init(_("Creating normalmap..."));
	
	if(normalmap(drawable,FALSE)==-1)
		status=GIMP_PDB_EXECUTION_ERROR;
	
	if(run_mode!=GIMP_RUN_NONINTERACTIVE)
		gimp_displays_flush();
	
	if(run_mode==GIMP_RUN_INTERACTIVE)
		gimp_set_data("plug_in_normalmap",&nmapvals,sizeof(nmapvals));
	
	values[0].data.d_status=status;
	
	gimp_drawable_detach(drawable);
}

#ifndef min
#define min(a,b)  ((a)<(b) ? (a) : (b))
#endif
#ifndef max
#define max(a,b)  ((a)>(b) ? (a) : (b))
#endif

static inline void NORMALIZE(float *v)
{
	float len=sqrtf(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
	
	if(len>1e-04f)
	{
		len=1.0f/len;
		v[0]*=len;
		v[1]*=len;
		v[2]*=len;
	}
	else
		v[0]=v[1]=v[2]=0.0f;
}

static void preview_do_row(gint row, gint width, guchar *even, guchar *odd,
                           guchar *src)
{
   gint    x;
  
   guchar *p0=even;
   guchar *p1=odd;
  
   gdouble r,g,b,a;
   gdouble c0,c1;
  
   for(x=0;x<width;x++) 
   {
      if(preview.bpp==4)
      {
         r=((gdouble)src[x*4+0])*oneover255;
         g=((gdouble)src[x*4+1])*oneover255;
         b=((gdouble)src[x*4+2])*oneover255;
         a=((gdouble)src[x*4+3])*oneover255;
      }
      else if(preview.bpp==3)
      {
         r=((gdouble)src[x*3+0])*oneover255;
         g=((gdouble)src[x*3+1])*oneover255;
         b=((gdouble)src[x*3+2])*oneover255;
         a=1.0;
      }
      else
      {
         r=((gdouble)src[x*preview.bpp+0])*oneover255;
         g=b=r;
         if(preview.bpp==2)
            a=((gdouble)src[x*preview.bpp+1])*oneover255;
         else
            a=1.0;
      }
      
      if((x/GIMP_CHECK_SIZE)&1) 
      {
         c0=GIMP_CHECK_LIGHT;
         c1=GIMP_CHECK_DARK;
      } 
      else 
      {
         c0=GIMP_CHECK_DARK;
         c1=GIMP_CHECK_LIGHT;
      }
      
      *p0++=(c0+(r-c0)*a)*255.0;
      *p0++=(c0+(g-c0)*a)*255.0;
      *p0++=(c0+(b-c0)*a)*255.0;
      
      *p1++=(c1+(r-c1)*a)*255.0;
      *p1++=(c1+(g-c1)*a)*255.0;
      *p1++=(c1+(b-c1)*a)*255.0;
   }
  
   if ((row / GIMP_CHECK_SIZE) & 1)
      gtk_preview_draw_row(GTK_PREVIEW(preview.widget),(guchar*)odd,0,row,width); 
   else
      gtk_preview_draw_row(GTK_PREVIEW(preview.widget),(guchar*)even,0,row,width);
}

typedef struct
{
   int x,y;
   float w;
} kernel_element;

static void make_kernel(kernel_element *k, float *weights, int size)
{
   int x,y,idx;
   
   for(y=0;y<size;++y)
   {
      for(x=0;x<size;++x)
      {
         idx=x+y*size;
         k[idx].x=x-(size/2);
         k[idx].y=(size/2)-y;
         k[idx].w=weights[idx];
      }
   }
}

static void rotate_array(float *dst, float *src, int size)
{
   int x,y,newx,newy;
   
   for(y=0;y<size;++y)
   {
      for(x=0;x<size;++x)
      {
         newy=size-x-1;
         newx=y;
         dst[newx+newy*size]=src[x+y*size];
      }
   }
}

static void scale_image(unsigned char *dst, int dw, int dh,
                        unsigned char *src, int sw, int sh,
                        int bpp)
{
   int n,x,y;
   int hm1=sh-1;
   int wm1=sw-1;
   int xfl,yfl;
   float xsrc,ysrc;
   float dx,dy,val;
   unsigned char *s,*d;
   int rowbytes=sw*bpp;
   
#define LERP(a,b,c) ((a)*(1.0f-(c))+((b)*(c)))
      
   d=dst;
      
   for(y=0;y<dh;y++)
   {
      ysrc=((float)y/(float)dh)*(float)hm1;
      yfl=(int)ysrc;
      dy=ysrc-(float)yfl;
      for(x=0;x<dw;x++)
      {
         xsrc=((float)x/(float)dw)*(float)wm1;
         xfl=(int)xsrc;
         dx=xsrc-(float)xfl;
         
         s=src+((yfl*rowbytes)+(xfl*bpp));
         for(n=0;n<bpp;++n)
         {
            val=LERP(LERP(s[0],s[bpp],dx),
                     LERP(s[rowbytes],s[rowbytes+bpp],dx),
                     dy);
            *d++=(unsigned char)val;
            s++;
         }
      }
   }
#undef LERP      
}

static gint32 normalmap(GimpDrawable *drawable, gboolean preview_mode)
{
	gint x,y;
	gint width,height,bpp,rowbytes;
	guchar *d,*dst,*s,*src,*tmp;
   guchar *even=0,*odd=0;
   float *heights;
	float val,du,dv,n[3],weight;
   float rgb_bias[3];
   int i,num_elements=0;
   kernel_element *kernel_du=0;
   kernel_element *kernel_dv=0;
	GimpPixelRgn src_rgn,dst_rgn;
   GdkCursor *cursor;
   
   if(nmapvals.filter<0||nmapvals.filter>=MAX_FILTER_TYPE)
      nmapvals.filter=FILTER_NONE;
   if(drawable->bpp!=4) nmapvals.height_source=0;
   if(drawable->bpp!=4&&(nmapvals.dudv==DUDV_16BIT_SIGNED||
                         nmapvals.dudv==DUDV_16BIT_UNSIGNED))
      nmapvals.dudv=DUDV_NONE;
	
	width=drawable->width;
	height=drawable->height;
	bpp=drawable->bpp;
	rowbytes=width*bpp;

   if(preview_mode)
   {
      even=g_malloc(preview.width*bpp);
      odd=g_malloc(preview.width*bpp);
   }
   
   //if(preview_mode)
   //   dst=g_malloc(preview.width*preview.height*bpp);
   //else
   dst=g_malloc(width*height*bpp);
   if(dst==0)
   {
      g_message(_("Memory allocation error!"));
      return(-1);
   }

   src=g_malloc(width*height*bpp);
	if(src==0)
	{
		g_message(_("Memory allocation error!"));
		return(-1);
	}
   
   //if(preview_mode)
   //   heights=g_new(float,preview.width*preview.height);
   //else
   heights=g_new(float,width*height);
   if(heights==0)
   {
      g_message(_("Memory allocation error!"));
      return(-1);
   }
   
   switch(nmapvals.filter)
   {
      case FILTER_NONE:
         num_elements=2;
         kernel_du=(kernel_element*)g_malloc(2*sizeof(kernel_element));
         kernel_dv=(kernel_element*)g_malloc(2*sizeof(kernel_element));
         
         kernel_du[0].x=-1; kernel_du[0].y= 0; kernel_du[0].w=-0.5f;
         kernel_du[1].x= 1; kernel_du[1].y= 0; kernel_du[1].w= 0.5f;
         
         kernel_dv[0].x= 0; kernel_dv[0].y= 1; kernel_dv[0].w= 0.5f;
         kernel_dv[1].x= 0; kernel_dv[1].y=-1; kernel_dv[1].w=-0.5f;
         
         break;
      case FILTER_SOBEL:
         num_elements=6;
         kernel_du=(kernel_element*)g_malloc(6*sizeof(kernel_element));
         kernel_dv=(kernel_element*)g_malloc(6*sizeof(kernel_element));
      
         kernel_du[0].x=-1; kernel_du[0].y= 1; kernel_du[0].w=-1.0f;
         kernel_du[1].x=-1; kernel_du[1].y= 0; kernel_du[1].w=-2.0f;
         kernel_du[2].x=-1; kernel_du[2].y=-1; kernel_du[2].w=-1.0f;
         kernel_du[3].x= 1; kernel_du[3].y= 1; kernel_du[3].w= 1.0f;
         kernel_du[4].x= 1; kernel_du[4].y= 0; kernel_du[4].w= 2.0f;
         kernel_du[5].x= 1; kernel_du[5].y=-1; kernel_du[5].w= 1.0f;
         
         kernel_dv[0].x=-1; kernel_dv[0].y= 1; kernel_dv[0].w= 1.0f;
         kernel_dv[1].x= 0; kernel_dv[1].y= 1; kernel_dv[1].w= 2.0f;
         kernel_dv[2].x= 1; kernel_dv[2].y= 1; kernel_dv[2].w= 1.0f;
         kernel_dv[3].x=-1; kernel_dv[3].y=-1; kernel_dv[3].w=-1.0f;
         kernel_dv[4].x= 0; kernel_dv[4].y=-1; kernel_dv[4].w=-2.0f;
         kernel_dv[5].x= 1; kernel_dv[5].y=-1; kernel_dv[5].w=-1.0f;
         
         break;
      case FILTER_PREWITT:
         num_elements=6;
         kernel_du=(kernel_element*)g_malloc(6*sizeof(kernel_element));
         kernel_dv=(kernel_element*)g_malloc(6*sizeof(kernel_element));
      
         kernel_du[0].x=-1; kernel_du[0].y= 1; kernel_du[0].w=-1.0f;
         kernel_du[1].x=-1; kernel_du[1].y= 0; kernel_du[1].w=-1.0f;
         kernel_du[2].x=-1; kernel_du[2].y=-1; kernel_du[2].w=-1.0f;
         kernel_du[3].x= 1; kernel_du[3].y= 1; kernel_du[3].w= 1.0f;
         kernel_du[4].x= 1; kernel_du[4].y= 0; kernel_du[4].w= 1.0f;
         kernel_du[5].x= 1; kernel_du[5].y=-1; kernel_du[5].w= 1.0f;
         
         kernel_dv[0].x=-1; kernel_dv[0].y= 1; kernel_dv[0].w= 1.0f;
         kernel_dv[1].x= 0; kernel_dv[1].y= 1; kernel_dv[1].w= 1.0f;
         kernel_dv[2].x= 1; kernel_dv[2].y= 1; kernel_dv[2].w= 1.0f;
         kernel_dv[3].x=-1; kernel_dv[3].y=-1; kernel_dv[3].w=-1.0f;
         kernel_dv[4].x= 0; kernel_dv[4].y=-1; kernel_dv[4].w=-1.0f;
         kernel_dv[5].x= 1; kernel_dv[5].y=-1; kernel_dv[5].w=-1.0f;
         
         break;      
      case FILTER_3x3:
         num_elements=6;
         kernel_du=(kernel_element*)g_malloc(6*sizeof(kernel_element));
         kernel_dv=(kernel_element*)g_malloc(6*sizeof(kernel_element));
      
         weight=1.0f/6.0f;
      
         kernel_du[0].x=-1; kernel_du[0].y= 1; kernel_du[0].w=-weight;
         kernel_du[1].x=-1; kernel_du[1].y= 0; kernel_du[1].w=-weight;
         kernel_du[2].x=-1; kernel_du[2].y=-1; kernel_du[2].w=-weight;
         kernel_du[3].x= 1; kernel_du[3].y= 1; kernel_du[3].w= weight;
         kernel_du[4].x= 1; kernel_du[4].y= 0; kernel_du[4].w= weight;
         kernel_du[5].x= 1; kernel_du[5].y=-1; kernel_du[5].w= weight;
         
         kernel_dv[0].x=-1; kernel_dv[0].y= 1; kernel_dv[0].w= weight;
         kernel_dv[1].x= 0; kernel_dv[1].y= 1; kernel_dv[1].w= weight;
         kernel_dv[2].x= 1; kernel_dv[2].y= 1; kernel_dv[2].w= weight;
         kernel_dv[3].x=-1; kernel_dv[3].y=-1; kernel_dv[3].w=-weight;
         kernel_dv[4].x= 0; kernel_dv[4].y=-1; kernel_dv[4].w=-weight;
         kernel_dv[5].x= 1; kernel_dv[5].y=-1; kernel_dv[5].w=-weight;
         break;
      case FILTER_5x5:
      {
         int n;
         float usum=0,vsum=0;
         float wt22=1.0f/16.0f;
         float wt12=1.0f/10.0f;
         float wt02=1.0f/8.0f;
         float wt11=1.0f/2.8f;
         num_elements=20;
         kernel_du=(kernel_element*)g_malloc(20*sizeof(kernel_element));
         kernel_dv=(kernel_element*)g_malloc(20*sizeof(kernel_element));
         
         kernel_du[0 ].x=-2; kernel_du[0 ].y= 2; kernel_du[0 ].w=-wt22;
         kernel_du[1 ].x=-1; kernel_du[1 ].y= 2; kernel_du[1 ].w=-wt12;
         kernel_du[2 ].x= 1; kernel_du[2 ].y= 2; kernel_du[2 ].w= wt12;
         kernel_du[3 ].x= 2; kernel_du[3 ].y= 2; kernel_du[3 ].w= wt22;
         kernel_du[4 ].x=-2; kernel_du[4 ].y= 1; kernel_du[4 ].w=-wt12;
         kernel_du[5 ].x=-1; kernel_du[5 ].y= 1; kernel_du[5 ].w=-wt11;
         kernel_du[6 ].x= 1; kernel_du[6 ].y= 1; kernel_du[6 ].w= wt11;
         kernel_du[7 ].x= 2; kernel_du[7 ].y= 1; kernel_du[7 ].w= wt12;
         kernel_du[8 ].x=-2; kernel_du[8 ].y= 0; kernel_du[8 ].w=-wt02;
         kernel_du[9 ].x=-1; kernel_du[9 ].y= 0; kernel_du[9 ].w=-0.5f;
         kernel_du[10].x= 1; kernel_du[10].y= 0; kernel_du[10].w= 0.5f;
         kernel_du[11].x= 2; kernel_du[11].y= 0; kernel_du[11].w= wt02;
         kernel_du[12].x=-2; kernel_du[12].y=-1; kernel_du[12].w=-wt12;
         kernel_du[13].x=-1; kernel_du[13].y=-1; kernel_du[13].w=-wt11;
         kernel_du[14].x= 1; kernel_du[14].y=-1; kernel_du[14].w= wt11;
         kernel_du[15].x= 2; kernel_du[15].y=-1; kernel_du[15].w= wt12;
         kernel_du[16].x=-2; kernel_du[16].y=-2; kernel_du[16].w=-wt22;
         kernel_du[17].x=-1; kernel_du[17].y=-2; kernel_du[17].w=-wt12;
         kernel_du[18].x= 1; kernel_du[18].y=-2; kernel_du[18].w= wt12;
         kernel_du[19].x= 2; kernel_du[19].y=-2; kernel_du[19].w= wt22;
         
         kernel_dv[0 ].x=-2; kernel_dv[0 ].y= 2; kernel_dv[0 ].w= wt22;
         kernel_dv[1 ].x=-1; kernel_dv[1 ].y= 2; kernel_dv[1 ].w= wt12;
         kernel_dv[2 ].x= 0; kernel_dv[2 ].y= 2; kernel_dv[2 ].w= 0.25f;
         kernel_dv[3 ].x= 1; kernel_dv[3 ].y= 2; kernel_dv[3 ].w= wt12;
         kernel_dv[4 ].x= 2; kernel_dv[4 ].y= 2; kernel_dv[4 ].w= wt22;
         kernel_dv[5 ].x=-2; kernel_dv[5 ].y= 1; kernel_dv[5 ].w= wt12;
         kernel_dv[6 ].x=-1; kernel_dv[6 ].y= 1; kernel_dv[6 ].w= wt11;
         kernel_dv[7 ].x= 0; kernel_dv[7 ].y= 1; kernel_dv[7 ].w= 0.5f;
         kernel_dv[8 ].x= 1; kernel_dv[8 ].y= 1; kernel_dv[8 ].w= wt11;
         kernel_dv[9 ].x= 2; kernel_dv[9 ].y= 1; kernel_dv[9 ].w= wt22;
         kernel_dv[10].x=-2; kernel_dv[10].y=-1; kernel_dv[10].w=-wt22;
         kernel_dv[11].x=-1; kernel_dv[11].y=-1; kernel_dv[11].w=-wt11;
         kernel_dv[12].x= 0; kernel_dv[12].y=-1; kernel_dv[12].w=-0.5f;
         kernel_dv[13].x= 1; kernel_dv[13].y=-1; kernel_dv[13].w=-wt11;
         kernel_dv[14].x= 2; kernel_dv[14].y=-1; kernel_dv[14].w=-wt12;
         kernel_dv[15].x=-2; kernel_dv[15].y=-2; kernel_dv[15].w=-wt22;
         kernel_dv[16].x=-1; kernel_dv[16].y=-2; kernel_dv[16].w=-wt12;
         kernel_dv[17].x= 0; kernel_dv[17].y=-2; kernel_dv[17].w=-0.25f;
         kernel_dv[18].x= 1; kernel_dv[18].y=-2; kernel_dv[18].w=-wt12;
         kernel_dv[19].x= 2; kernel_dv[19].y=-2; kernel_dv[19].w=-wt22;

         for(n=0;n<20;n++)
         {
            usum+=fabsf(kernel_du[n].w);
            vsum+=fabsf(kernel_dv[n].w);
         }
         for(n=0;n<20;n++)
         {
            kernel_du[n].w/=usum;
            kernel_dv[n].w/=vsum;
         }
         
         break;
      }
      case FILTER_7x7:
      {
         float du_weights[]=
         {
            -1,-2,-3, 0, 3, 2, 1,
            -2,-3,-4, 0, 4, 3, 2,
            -3,-4,-5, 0, 5, 4, 3,
            -4,-5,-6, 0, 6, 5, 4,
            -3,-4,-5, 0, 5, 4, 3,
            -2,-3,-4, 0, 4, 3, 2,
            -1,-2,-3, 0, 3, 2, 1   
         };
         float dv_weights[49];
         int n;
         float usum=0,vsum=0;
         
         num_elements=49;
         kernel_du=(kernel_element*)g_malloc(49*sizeof(kernel_element));
         kernel_dv=(kernel_element*)g_malloc(49*sizeof(kernel_element));
         
         make_kernel(kernel_du,du_weights,7);
         rotate_array(dv_weights,du_weights,7);
         make_kernel(kernel_dv,dv_weights,7);
         
         for(n=0;n<49;n++)
         {
            usum+=fabsf(kernel_du[n].w);
            vsum+=fabsf(kernel_dv[n].w);
         }
         for(n=0;n<49;n++)
         {
            kernel_du[n].w/=usum;
            kernel_dv[n].w/=vsum;
         }
         
         break;
      }
      case FILTER_9x9:
      {
         float du_weights[]=
         {
            -1,-2,-3,-4, 0, 4, 3, 2, 1,
            -2,-3,-4,-5, 0, 5, 4, 3, 2,
            -3,-4,-5,-6, 0, 6, 5, 4, 3,
            -4,-5,-6,-7, 0, 7, 6, 5, 4,
            -5,-6,-7,-8, 0, 8, 7, 6, 5,
            -4,-5,-6,-7, 0, 7, 6, 5, 4,
            -3,-4,-5,-6, 0, 6, 5, 4, 3,
            -2,-3,-4,-5, 0, 5, 4, 3, 2,
            -1,-2,-3,-4, 0, 4, 3, 2, 1     
         };
         float dv_weights[81];
         int n;
         float usum=0,vsum=0;
         
         num_elements=81;
         kernel_du=(kernel_element*)g_malloc(81*sizeof(kernel_element));
         kernel_dv=(kernel_element*)g_malloc(81*sizeof(kernel_element));
         
         make_kernel(kernel_du,du_weights,9);
         rotate_array(dv_weights,du_weights,9);
         make_kernel(kernel_dv,dv_weights,9);
         
         for(n=0;n<81;n++)
         {
            usum+=fabsf(kernel_du[n].w);
            vsum+=fabsf(kernel_dv[n].w);
         }
         for(n=0;n<81;n++)
         {
            kernel_du[n].w/=usum;
            kernel_dv[n].w/=vsum;
         }
         
         break;
      }
   }

   gimp_pixel_rgn_init(&src_rgn,drawable,0,0,width,height,0,0);
   gimp_pixel_rgn_get_rect(&src_rgn,src,0,0,width,height);
   /*
	if(preview_mode)
   {
      scale_image(dst,preview.width,preview.height,
                  src,width,height,bpp);
      
      width=preview.width;
      height=preview.height;
      rowbytes=preview.width*bpp;
      
      g_free(src);
      src=g_malloc(width*height*bpp);
      memcpy(src,dst,width*height*bpp);
   }
   */
   if(nmapvals.conversion==CONVERT_BIASED_RGB)
   {
      /* approximated average color of the image
       * scale to 16x16, accumulate the pixels and average */
      unsigned char *tmp;
      int sum[3];

      tmp=g_malloc(16*16*bpp);
      scale_image(tmp,16,16,src,width,height,bpp);

      sum[0]=sum[1]=sum[2]=0;
      
      s=tmp;
      for(y=0;y<16;y++)
      {
         for(x=0;x<16;x++)
         {
            sum[0]+=*s++;
            sum[1]+=*s++;
            sum[2]+=*s++;
            if(bpp==4) s++;
         }
      }
         
      rgb_bias[0]=(float)sum[0]/256.0f;
      rgb_bias[1]=(float)sum[1]/256.0f;
      rgb_bias[2]=(float)sum[2]/256.0f;
      
      g_free(tmp);
   }
   else
   {
      rgb_bias[0]=0;
      rgb_bias[1]=0;
      rgb_bias[2]=0;
   }

   if(nmapvals.conversion!=CONVERT_NORMALIZE_ONLY)
   {
      s=src;
      for(y=0;y<height;y++)
      {
         for(x=0;x<width;x++)
         {
            if(!nmapvals.height_source)
            {
               switch(nmapvals.conversion)
               {
                  case CONVERT_NONE:
                     val=(float)s[0]*0.3f+(float)s[1]*0.59f+(float)s[2]*0.11f;
                     break;
                  case CONVERT_BIASED_RGB:
                     val=(((float)s[0]-rgb_bias[0])*0.3f )+
                         (((float)s[1]-rgb_bias[1])*0.59f)+
                         (((float)s[2]-rgb_bias[2])*0.11f);
                     break;
                  case CONVERT_RED:
                     val=(float)s[0];
                     break;
                  case CONVERT_GREEN:
                     val=(float)s[1];
                     break;
                  case CONVERT_BLUE:
                     val=(float)s[2];
                     break;
                  case CONVERT_MAX_RGB:
                     val=(float)max(s[0],max(s[1],s[2]));
                     break;
                  case CONVERT_COLORSPACE:
                     val=(1.0f-((1.0f-((float)s[0]/255.0f))*
                                (1.0f-((float)s[1]/255.0f))*
                                (1.0f-((float)s[2]/255.0f))))*255.0f;
                     break;
                  default:
                     val=255.0f;
                     break;
               }
            }
            else
               val=(float)s[3];
         
            heights[x+y*width]=val*oneover255;
         
            s+=bpp;
         }
      }
   }

#define HEIGHT(x,y) \
	(heights[(max(0,min(width-1,(x))))+(max(0,min(height-1,(y))))*width])
#define HEIGHT_WRAP(x,y) \
   (heights[((x)<0 ? (width+(x)) : ((x)>=width ? ((x)-width) : (x)))+ \
            (((y)<0 ? (height+(y)) : ((y)>=height ? ((y)-height) : (y)))*width)])
   
   if(preview_mode)
   {
      cursor=gdk_cursor_new(GDK_WATCH);
      gdk_window_set_cursor(dialog->window,cursor);
      gdk_cursor_destroy(cursor);
   }
	
	for(y=0;y<height;y++)
	{
      while(gtk_events_pending())
         gtk_main_iteration();
      
		for(x=0;x<width;x++)
		{
			d=dst+((y*rowbytes)+(x*bpp));
         s=src+((y*rowbytes)+(x*bpp));

         if(nmapvals.conversion==CONVERT_NORMALIZE_ONLY)
         {
            n[0]=(((float)s[0]*oneover255)-0.5f)*2.0f;
            n[1]=(((float)s[1]*oneover255)-0.5f)*2.0f;
            n[2]=(((float)s[2]*oneover255)-0.5f)*2.0f;
         }
         else
         {
            du=0; dv=0;
            if(!nmapvals.wrap)
            {
               for(i=0;i<num_elements;++i)
                  du+=HEIGHT(x+kernel_du[i].x,
                             y+kernel_du[i].y)*kernel_du[i].w;
               for(i=0;i<num_elements;++i)
                  dv+=HEIGHT(x+kernel_dv[i].x,
                             y+kernel_dv[i].y)*kernel_dv[i].w;
            }
            else
            {
               for(i=0;i<num_elements;++i)
                  du+=HEIGHT_WRAP(x+kernel_du[i].x,
                                  y+kernel_du[i].y)*kernel_du[i].w;
               for(i=0;i<num_elements;++i)
                  dv+=HEIGHT_WRAP(x+kernel_dv[i].x,
                                  y+kernel_dv[i].y)*kernel_dv[i].w;
            }
            
            n[0]=-du*nmapvals.scale;
            n[1]=-dv*nmapvals.scale;
            n[2]=1.0f;
         }
         
         NORMALIZE(n);

         if(!nmapvals.dudv)
         {
            *d++=(unsigned char)((n[0]+1.0f)*127.5f);
            *d++=(unsigned char)((n[1]+1.0f)*127.5f);
            *d++=(unsigned char)((n[2]+1.0f)*127.5f);
         
            if(drawable->bpp==4)
            {
               switch(nmapvals.alpha)
               {
                  case 0:
                     *d++=s[3]; break;
                  case 1:
                     *d++=(unsigned char)(heights[x+y*width]*255.0f); break;
                  case 2:
                     *d++=0; break;
                  case 3:
                     *d++=255; break;
                  default:
                     *d++=s[3]; break;
               }
            }
         }
         else
         {
            if(nmapvals.dudv==DUDV_8BIT_SIGNED||
               nmapvals.dudv==DUDV_8BIT_UNSIGNED)
            {
               if(nmapvals.dudv==DUDV_8BIT_UNSIGNED)
               {
                  n[0]+=1.0f;
                  n[1]+=1.0f;
               }
               *d++=(unsigned char)(n[0]*127.5f);
               *d++=(unsigned char)(n[1]*127.5f);
               *d++=0;
               if(drawable->bpp==4) *d++=255;
            }
            else if(nmapvals.dudv==DUDV_16BIT_SIGNED||
                    nmapvals.dudv==DUDV_16BIT_UNSIGNED)
            {
               unsigned short *d16=(unsigned short*)d;
               if(nmapvals.dudv==DUDV_16BIT_UNSIGNED)
               {
                  n[0]+=1.0f;
                  n[1]+=1.0f;
               }
               *d16++=(unsigned short)(n[0]*32767.5f);
               *d16++=(unsigned short)(n[1]*32767.5f);
            }
         }
		}

		if(!preview_mode)
         gimp_progress_update((double)(y-1)/(double)(height-2));
	}
	
#undef HEIGHT
#undef HEIGHT_WRAP

   if(preview_mode)
   {
      update_3D_preview(width,height,bpp,dst);
      
      tmp=g_malloc(preview.width*preview.height*bpp);
      scale_image(tmp,preview.width,preview.height,dst,width,height,bpp);
      
      width=preview.width;
      height=preview.height;
      rowbytes=preview.width*bpp;
      
      for(y=0;y<height;++y)
         preview_do_row(y,width,even,odd,tmp+(y*rowbytes));
      
      gtk_widget_queue_draw(preview.widget);
      
      g_free(even);
      g_free(odd);
      g_free(tmp);
      
      gdk_window_set_cursor(dialog->window,0);
   }
   else
   {
      gimp_progress_update(100.0);
      
      gimp_pixel_rgn_init(&dst_rgn,drawable,0,0,width,height,1,1);
      gimp_pixel_rgn_set_rect(&dst_rgn,dst,0,0,width,height);
      
      gimp_drawable_flush(drawable);
      gimp_drawable_merge_shadow(drawable->id,1);
      gimp_drawable_update(drawable->id,0,0,width,height);
   }
   
	g_free(heights);
	g_free(src);
	g_free(dst);
	g_free(kernel_du);
   g_free(kernel_dv);
   
	return(0);
}

static void create_preview_widget(GimpDrawable *drawable)
{
   guchar *drawable_data;
   gint bpp;
   gint y;
   gint w=PREVIEW_SIZE;
   gint h=PREVIEW_SIZE;
   guchar *src;
   guchar *even,*odd;

   preview.widget=gtk_preview_new(GTK_PREVIEW_COLOR);

   bpp=0; /* Only returned */
  
   drawable_data=gimp_drawable_get_thumbnail_data(drawable->id,&w,&h,&bpp);

   if(w<1||h<1) return;

   preview.width=w;
   preview.height=h;
   preview.cache=drawable_data;
   preview.rowstride=w*bpp;
   preview.bpp=bpp;

   gtk_preview_size(GTK_PREVIEW(preview.widget),w,h);

   even=g_malloc(w*3);
   odd=g_malloc(w*3);
   src=drawable_data;

   for(y=0;y<h;y++)
      preview_do_row(y,w,even,odd,preview.cache+(y*w*bpp));

   g_free(even);
   g_free(odd);
}

static void ok_callback(GtkWidget *widget, gpointer data)
{
	runme=1;
   destroy_3D_preview();
	gtk_widget_destroy(GTK_WIDGET(data));
}

static int update_preview=0;

static gint idle_callback(gpointer data)
{
   if(!update_preview) return(1);

   update_preview=0;
   normalmap((GimpDrawable*)data,TRUE);

   return(1);
}

static void filter_type_selected(GtkWidget *widget, gpointer data)
{
   if(nmapvals.filter!=(gint)data)
   {
      nmapvals.filter=(gint)data;
      update_preview=1;
   }
}

static void scale_changed(GtkWidget *widget, gpointer data)
{
   nmapvals.scale=gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(data));
   update_preview=1;
}

static void wrap_mode_selected(GtkWidget *widget, gpointer data)
{
   if(nmapvals.wrap!=(gint)data)
   {
      nmapvals.wrap=(gint)data;
      update_preview=1;
   }
}

static void height_source_selected(GtkWidget *widget, gpointer data)
{
   GtkWidget *opt,*menu;

	if(nmapvals.height_source==(gint)data) return;
   
   nmapvals.height_source=(gint)data;
   
   opt=gtk_object_get_data(GTK_OBJECT(widget),"conversion_opt");
   if(!nmapvals.height_source)
   {
      gtk_widget_set_sensitive(opt,1);
   }
   else
   {
      nmapvals.conversion=CONVERT_NONE;
      menu=gtk_option_menu_get_menu(GTK_OPTION_MENU(opt));
      gtk_menu_set_active(GTK_MENU(menu),nmapvals.conversion);
      gtk_widget_set_sensitive(opt,0);
   }
      
   update_preview=1;
}

static void alpha_result_selected(GtkWidget *widget, gpointer data)
{
   if(nmapvals.alpha!=(gint)data)
   {
      nmapvals.alpha=(gint)data;
      update_preview=1;
   }
}

static void conversion_selected(GtkWidget *widget, gpointer data)
{
   if(nmapvals.conversion!=(gint)data)
   {
      nmapvals.conversion=(gint)data;
      update_preview=1;
   }
}

static void preview_clicked(GtkWidget *widget, gpointer data)
{
   GimpDrawable *drawable;
   if(!is_3D_preview_active())
   {
      drawable=gtk_object_get_data(GTK_OBJECT(widget),"drawable");
      show_3D_preview(drawable);
      update_preview=1;
   }
}

static void dudv_selected(GtkWidget *widget, gpointer data)
{
   GimpDrawable *drawable;
   GtkWidget *opt,*menu;
   
   if(nmapvals.dudv==(gint)data) return;
   
	nmapvals.dudv=(gint)data;
   
   drawable=gtk_object_get_data(GTK_OBJECT(widget),"drawable");
   opt=gtk_object_get_data(GTK_OBJECT(widget),"alpha_opt");

   if(nmapvals.dudv==DUDV_NONE)
   {
      if(drawable->bpp==4)
         gtk_widget_set_sensitive(opt,1);
   }
   else
   {
      nmapvals.alpha=0;
      menu=gtk_option_menu_get_menu(GTK_OPTION_MENU(opt));
      gtk_menu_set_active(GTK_MENU(menu),nmapvals.alpha);
      gtk_widget_set_sensitive(opt,0);
   }
      
   update_preview=1;
}

static gint normalmap_dialog(GimpDrawable *drawable)
{
   GtkWidget *hbox;
   GtkWidget *vbox;
   GtkWidget *frame;
   GtkWidget *abox;
   GtkWidget *btn;
   GtkWidget *table;
   GtkWidget *opt,*alpha_result_opt;
   GtkWidget *menu;
   GtkWidget *menuitem,*item_height_source[2];
   GtkWidget *label;
   GtkObject *adj;
   GtkWidget *spin;
		
	dialog=gimp_dialog_new(_("Normalmap"),"normalmap",
                          gimp_standard_help_func,0,
                          GTK_WIN_POS_MOUSE,0,1,0,
                          _("Ok"),ok_callback,
                          0,0,0,1,0,
                          _("Cancel"),gtk_widget_destroy,
                          0,1,0,0,1,
                          0);
	gtk_signal_connect(GTK_OBJECT(dialog),"destroy",
							 GTK_SIGNAL_FUNC(gtk_main_quit),
							 0);
   
   hbox=gtk_hbox_new(0,10);
   gtk_container_set_border_width(GTK_CONTAINER(hbox),10);
   gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),hbox,1,1,0);
   gtk_widget_show(hbox);
   
   vbox=gtk_vbox_new(0,10);
   gtk_box_pack_start(GTK_BOX(hbox),vbox,1,1,0);
   gtk_widget_show(vbox);
   
   frame=gtk_frame_new(_("Preview"));
   gtk_box_pack_start(GTK_BOX(vbox),frame,0,0,0);
   gtk_widget_show(frame);
   abox=gtk_alignment_new(0.5,0.5,0.0,0.0);
   gtk_container_set_border_width(GTK_CONTAINER(abox),5);
   gtk_container_add(GTK_CONTAINER(frame),abox);
   gtk_widget_show(abox);
   frame=gtk_frame_new(NULL);
   gtk_frame_set_shadow_type(GTK_FRAME(frame),GTK_SHADOW_IN);
   gtk_container_add(GTK_CONTAINER(abox),frame);
   gtk_widget_show(frame);
   create_preview_widget(drawable);
   gtk_container_add(GTK_CONTAINER(frame),preview.widget);
   normalmap(drawable,TRUE);
   gtk_widget_show(preview.widget);
   
   btn=gtk_button_new_with_label(_("3D Preview"));
   gtk_signal_connect(GTK_OBJECT(btn),"clicked",
                      GTK_SIGNAL_FUNC(preview_clicked),0);
   gtk_object_set_data(GTK_OBJECT(btn),"drawable",drawable);
   gtk_box_pack_start(GTK_BOX(vbox),btn,0,0,0);
   gtk_widget_show(btn);
   
	table=gtk_table_new(7,2,0);
	gtk_widget_show(table);
	gtk_box_pack_start(GTK_BOX(hbox),table,1,1,0);
	gtk_table_set_row_spacings(GTK_TABLE(table),10);
	gtk_table_set_col_spacings(GTK_TABLE(table),10);
	
	label=gtk_label_new(_("Filter:"));
	gtk_widget_show(label);
	gtk_table_attach(GTK_TABLE(table),label,0,1,0,1,
						  (GtkAttachOptions)(GTK_FILL),
						  (GtkAttachOptions)(0),0,0);
	gtk_misc_set_alignment(GTK_MISC(label),0,0.5);
	
	opt=gtk_option_menu_new();
	gtk_widget_show(opt);
	gtk_table_attach(GTK_TABLE(table),opt,1,2,0,1,
						  (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),
						  (GtkAttachOptions)(GTK_EXPAND),0,0);
	
	menu=gtk_menu_new();
	
	menuitem=gtk_menu_item_new_with_label(_("4 sample"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(filter_type_selected),
							 (gpointer)FILTER_NONE);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	menuitem=gtk_menu_item_new_with_label("Sobel");
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(filter_type_selected),
							 (gpointer)FILTER_SOBEL);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	menuitem=gtk_menu_item_new_with_label("Prewitt");
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(filter_type_selected),
							 (gpointer)FILTER_PREWITT);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	menuitem=gtk_menu_item_new_with_label("3x3");
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(filter_type_selected),
							 (gpointer)FILTER_3x3);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	menuitem=gtk_menu_item_new_with_label("5x5");
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(filter_type_selected),
							 (gpointer)FILTER_5x5);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	menuitem=gtk_menu_item_new_with_label("7x7");
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(filter_type_selected),
							 (gpointer)FILTER_7x7);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	menuitem=gtk_menu_item_new_with_label("9x9");
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(filter_type_selected),
							 (gpointer)FILTER_9x9);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	gtk_menu_set_active(GTK_MENU(menu),nmapvals.filter);
	gtk_option_menu_set_menu(GTK_OPTION_MENU(opt),menu);

	label=gtk_label_new(_("Scale:"));
	gtk_widget_show(label);
	gtk_table_attach(GTK_TABLE(table),label,0,1,1,2,
						  (GtkAttachOptions)(GTK_FILL),
						  (GtkAttachOptions)(0),0,0);
	gtk_misc_set_alignment(GTK_MISC(label),0,0.5);

   adj=gtk_adjustment_new(nmapvals.scale,-100,100,1,5,5);
   spin=gtk_spin_button_new(GTK_ADJUSTMENT(adj),1,5);
   gtk_signal_connect(GTK_OBJECT(adj),"value_changed",
                      GTK_SIGNAL_FUNC(scale_changed),spin);
	gtk_table_attach(GTK_TABLE(table),spin,1,2,1,2,
						  (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),
						  (GtkAttachOptions)(GTK_EXPAND),0,0);
   gtk_widget_show(spin);
   
	label=gtk_label_new(_("Wrap:"));
	gtk_widget_show(label);
	gtk_table_attach(GTK_TABLE(table),label,0,1,2,3,
						  (GtkAttachOptions)(GTK_FILL),
						  (GtkAttachOptions)(0),0,0);
	gtk_misc_set_alignment(GTK_MISC(label),0,0.5);
	
	opt=gtk_option_menu_new();
	gtk_widget_show(opt);
	gtk_table_attach(GTK_TABLE(table),opt,1,2,2,3,
						  (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),
						  (GtkAttachOptions)(GTK_EXPAND),0,0);
	
	menu=gtk_menu_new();
	
	menuitem=gtk_menu_item_new_with_label(_("No"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(wrap_mode_selected),
							 (gpointer)0);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	menuitem=gtk_menu_item_new_with_label(_("Yes"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(wrap_mode_selected),
							 (gpointer)1);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	gtk_menu_set_active(GTK_MENU(menu),nmapvals.wrap);
	gtk_option_menu_set_menu(GTK_OPTION_MENU(opt),menu);

	label=gtk_label_new(_("Height source:"));
	gtk_widget_show(label);
	gtk_table_attach(GTK_TABLE(table),label,0,1,3,4,
						  (GtkAttachOptions)(GTK_FILL),
						  (GtkAttachOptions)(0),0,0);
	gtk_misc_set_alignment(GTK_MISC(label),0,0.5);
	
	opt=gtk_option_menu_new();
	gtk_widget_show(opt);
	gtk_table_attach(GTK_TABLE(table),opt,1,2,3,4,
						  (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),
						  (GtkAttachOptions)(GTK_EXPAND),0,0);
	
	menu=gtk_menu_new();
   
   if(drawable->bpp!=4)
      nmapvals.height_source=0;
	
	menuitem=item_height_source[0]=gtk_menu_item_new_with_label(_("Average RGB"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(height_source_selected),
							 (gpointer)0);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=item_height_source[1]=gtk_menu_item_new_with_label(_("Alpha"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(height_source_selected),
                      (gpointer)1);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);

	gtk_menu_set_active(GTK_MENU(menu),nmapvals.height_source);
   gtk_option_menu_set_menu(GTK_OPTION_MENU(opt),menu);
   
   if(drawable->bpp!=4)
      gtk_widget_set_sensitive(opt,0);
   
	label=gtk_label_new(_("Alpha channel:"));
	gtk_widget_show(label);
	gtk_table_attach(GTK_TABLE(table),label,0,1,4,5,
						  (GtkAttachOptions)(GTK_FILL),
						  (GtkAttachOptions)(0),0,0);
	gtk_misc_set_alignment(GTK_MISC(label),0,0.5);
	
	opt=alpha_result_opt=gtk_option_menu_new();
	gtk_widget_show(opt);
	gtk_table_attach(GTK_TABLE(table),opt,1,2,4,5,
						  (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),
						  (GtkAttachOptions)(GTK_EXPAND),0,0);
   
   if(drawable->bpp!=4)
      nmapvals.alpha=0;
      
	menu=gtk_menu_new();
	
	menuitem=gtk_menu_item_new_with_label(_("Unchanged"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(alpha_result_selected),
							 (gpointer)0);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=gtk_menu_item_new_with_label(_("Height"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(alpha_result_selected),
                      (gpointer)1);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=gtk_menu_item_new_with_label(_("Set to 0"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(alpha_result_selected),
                      (gpointer)2);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=gtk_menu_item_new_with_label(_("Set to 1"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(alpha_result_selected),
                      (gpointer)3);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);

   gtk_menu_set_active(GTK_MENU(menu),nmapvals.alpha);
	gtk_option_menu_set_menu(GTK_OPTION_MENU(opt),menu);
   
   if(drawable->bpp!=4||nmapvals.dudv!=DUDV_NONE)
      gtk_widget_set_sensitive(opt,0);

	label=gtk_label_new(_("Conversion:"));
	gtk_widget_show(label);
	gtk_table_attach(GTK_TABLE(table),label,0,1,5,6,
						  (GtkAttachOptions)(GTK_FILL),
						  (GtkAttachOptions)(0),0,0);
	gtk_misc_set_alignment(GTK_MISC(label),0,0.5);
	
	opt=gtk_option_menu_new();
	gtk_widget_show(opt);
	gtk_table_attach(GTK_TABLE(table),opt,1,2,5,6,
						  (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),
						  (GtkAttachOptions)(GTK_EXPAND),0,0);
   
   gtk_object_set_data(GTK_OBJECT(item_height_source[0]),"conversion_opt",opt);
   gtk_object_set_data(GTK_OBJECT(item_height_source[1]),"conversion_opt",opt);
   
	menu=gtk_menu_new();
	
	menuitem=gtk_menu_item_new_with_label(_("None"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(conversion_selected),
							 (gpointer)CONVERT_NONE);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=gtk_menu_item_new_with_label(_("Biased RGB"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(conversion_selected),
                      (gpointer)CONVERT_BIASED_RGB);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=gtk_menu_item_new_with_label(_("Red"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(conversion_selected),
                      (gpointer)CONVERT_RED);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=gtk_menu_item_new_with_label(_("Green"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(conversion_selected),
                      (gpointer)CONVERT_GREEN);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=gtk_menu_item_new_with_label(_("Blue"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(conversion_selected),
                      (gpointer)CONVERT_BLUE);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=gtk_menu_item_new_with_label(_("Max RGB"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(conversion_selected),
                      (gpointer)CONVERT_MAX_RGB);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=gtk_menu_item_new_with_label(_("Colorspace"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(conversion_selected),
                      (gpointer)CONVERT_COLORSPACE);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);
   menuitem=gtk_menu_item_new_with_label(_("Normalize only"));
   gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
                      GTK_SIGNAL_FUNC(conversion_selected),
                      (gpointer)CONVERT_NORMALIZE_ONLY);
   gtk_widget_show(menuitem);
   gtk_menu_append(GTK_MENU(menu),menuitem);

   gtk_menu_set_active(GTK_MENU(menu),nmapvals.conversion);
	gtk_option_menu_set_menu(GTK_OPTION_MENU(opt),menu);
   
   if(nmapvals.height_source)
      gtk_widget_set_sensitive(opt,0);
 
	label=gtk_label_new(_("DU/DV map:"));
	gtk_widget_show(label);
	gtk_table_attach(GTK_TABLE(table),label,0,1,6,7,
						  (GtkAttachOptions)(GTK_FILL),
						  (GtkAttachOptions)(0),0,0);
	gtk_misc_set_alignment(GTK_MISC(label),0,0.5);
	
	opt=gtk_option_menu_new();
	gtk_widget_show(opt);
	gtk_table_attach(GTK_TABLE(table),opt,1,2,6,7,
						  (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),
						  (GtkAttachOptions)(GTK_EXPAND),0,0);
   
   if(drawable->bpp!=4&&(nmapvals.dudv==DUDV_16BIT_SIGNED||
                         nmapvals.dudv==DUDV_16BIT_UNSIGNED))
      nmapvals.dudv=DUDV_NONE;
   
	menu=gtk_menu_new();
   	
	menuitem=gtk_menu_item_new_with_label(_("None"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(dudv_selected),
							 (gpointer)DUDV_NONE);
   gtk_object_set_data(GTK_OBJECT(menuitem),"drawable",drawable);
   gtk_object_set_data(GTK_OBJECT(menuitem),"alpha_opt",alpha_result_opt);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	menuitem=gtk_menu_item_new_with_label(_("8 bits"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(dudv_selected),
							 (gpointer)DUDV_8BIT_SIGNED);
   gtk_object_set_data(GTK_OBJECT(menuitem),"drawable",drawable);
   gtk_object_set_data(GTK_OBJECT(menuitem),"alpha_opt",alpha_result_opt);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	menuitem=gtk_menu_item_new_with_label(_("8 bits (unsigned)"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(dudv_selected),
							 (gpointer)DUDV_8BIT_UNSIGNED);
   gtk_object_set_data(GTK_OBJECT(menuitem),"drawable",drawable);
   gtk_object_set_data(GTK_OBJECT(menuitem),"alpha_opt",alpha_result_opt);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
	menuitem=gtk_menu_item_new_with_label(_("16 bits"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(dudv_selected),
							 (gpointer)DUDV_16BIT_SIGNED);
   gtk_object_set_data(GTK_OBJECT(menuitem),"drawable",drawable);
   gtk_object_set_data(GTK_OBJECT(menuitem),"alpha_opt",alpha_result_opt);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
   if(drawable->bpp!=4)
      gtk_widget_set_sensitive(menuitem,0);
	menuitem=gtk_menu_item_new_with_label(_("16 bits (unsigned)"));
	gtk_signal_connect(GTK_OBJECT(menuitem),"activate",
							 GTK_SIGNAL_FUNC(dudv_selected),
							 (gpointer)DUDV_16BIT_UNSIGNED);
   gtk_object_set_data(GTK_OBJECT(menuitem),"drawable",drawable);
   gtk_object_set_data(GTK_OBJECT(menuitem),"alpha_opt",alpha_result_opt);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(menu),menuitem);
   if(drawable->bpp!=4)
      gtk_widget_set_sensitive(menuitem,0);

   gtk_menu_set_active(GTK_MENU(menu),nmapvals.dudv);
	gtk_option_menu_set_menu(GTK_OPTION_MENU(opt),menu);
   
	gtk_widget_show(dialog);
   
   update_preview=0;
   gtk_timeout_add(100,idle_callback,drawable);
	
	gtk_main();
	gdk_flush();
	
	return(runme);
}
